@@ -20,7 +20,7 @@
  * The implementation is heavily inspired by pg_stat_statements
  *
  * Copyright (c) 2014,2017 Ronan Dunklau
- * Copyright (c) 2018, The Powa-Team
+ * Copyright (c) 2018-2022, The Powa-Team
  *-------------------------------------------------------------------------
  */
 #include <limits.h>
@@ -42,6 +42,9 @@
 #include "catalog/pg_operator.h"
 #include "catalog/pg_type.h"
 #include "commands/dbcommands.h"
+#if PG_VERSION_NUM >= 150000
+#include "common/pg_prng.h"
+#endif
 #include "fmgr.h"
 #include "funcapi.h"
 #include "mb/pg_wchar.h"
@@ -54,7 +57,13 @@
 #include "parser/analyze.h"
 #include "parser/parse_node.h"
 #include "parser/parsetree.h"
+#if PG_VERSION_NUM >= 150000
 #include "postmaster/autovacuum.h"
+#endif
+#include "postmaster/postmaster.h"
+#if PG_VERSION_NUM >= 150000
+#include "replication/walsender.h"
+#endif
 #include "storage/ipc.h"
 #include "storage/lwlock.h"
 #if PG_VERSION_NUM >= 100000
@@ -106,18 +115,17 @@ typedef enum pgqsVersion
 
 /*---- Function declarations ----*/
 
-void		_PG_init(void);
-void		_PG_fini(void);
+extern PGDLLEXPORT void		_PG_init(void);
 
-extern Datum pg_qualstats_reset(PG_FUNCTION_ARGS);
-extern Datum pg_qualstats(PG_FUNCTION_ARGS);
-extern Datum pg_qualstats_2_0(PG_FUNCTION_ARGS);
-extern Datum pg_qualstats_names(PG_FUNCTION_ARGS);
-extern Datum pg_qualstats_names_2_0(PG_FUNCTION_ARGS);
+extern PGDLLEXPORT Datum pg_qualstats_reset(PG_FUNCTION_ARGS);
+extern PGDLLEXPORT Datum pg_qualstats(PG_FUNCTION_ARGS);
+extern PGDLLEXPORT Datum pg_qualstats_2_0(PG_FUNCTION_ARGS);
+extern PGDLLEXPORT Datum pg_qualstats_names(PG_FUNCTION_ARGS);
+extern PGDLLEXPORT Datum pg_qualstats_names_2_0(PG_FUNCTION_ARGS);
 static Datum pg_qualstats_common(PG_FUNCTION_ARGS, pgqsVersion api_version,
 								 bool include_names);
-extern Datum pg_qualstats_example_query(PG_FUNCTION_ARGS);
-extern Datum pg_qualstats_example_queries(PG_FUNCTION_ARGS);
+extern PGDLLEXPORT Datum pg_qualstats_example_query(PG_FUNCTION_ARGS);
+extern PGDLLEXPORT Datum pg_qualstats_example_queries(PG_FUNCTION_ARGS);
 
 PG_FUNCTION_INFO_V1(pg_qualstats_reset);
 PG_FUNCTION_INFO_V1(pg_qualstats);
@@ -128,6 +136,9 @@ PG_FUNCTION_INFO_V1(pg_qualstats_example_query);
 PG_FUNCTION_INFO_V1(pg_qualstats_example_queries);
 
 static void pgqs_backend_mode_startup(void);
+#if PG_VERSION_NUM >= 150000
+static void pgqs_shmem_request(void);
+#endif
 static void pgqs_shmem_startup(void);
 static void pgqs_ExecutorStart(QueryDesc *queryDesc, int eflags);
 static void pgqs_ExecutorRun(QueryDesc *queryDesc,
@@ -148,6 +159,9 @@ static ExecutorStart_hook_type prev_ExecutorStart = NULL;
 static ExecutorRun_hook_type prev_ExecutorRun = NULL;
 static ExecutorFinish_hook_type prev_ExecutorFinish = NULL;
 static ExecutorEnd_hook_type prev_ExecutorEnd = NULL;
+#if PG_VERSION_NUM >= 150000
+static shmem_request_hook_type prev_shmem_request_hook = NULL;
+#endif
 static shmem_startup_hook_type prev_shmem_startup_hook = NULL;
 
 static uint32 pgqs_hash_fn(const void *key, Size keysize);
@@ -349,6 +363,10 @@ _PG_init(void)
 	else
 	{
 		pgqs_backend = false;
+#if PG_VERSION_NUM >= 150000
+		prev_shmem_request_hook = shmem_request_hook;
+		shmem_request_hook = pgqs_shmem_request;
+#endif
 		prev_shmem_startup_hook = shmem_startup_hook;
 		shmem_startup_hook = pgqs_shmem_startup;
 	}
@@ -466,28 +484,19 @@ _PG_init(void)
 
 	if (!pgqs_backend)
 	{
+#if PG_VERSION_NUM < 150000
 		RequestAddinShmemSpace(pgqs_memsize());
 #if PG_VERSION_NUM >= 90600
 		RequestNamedLWLockTranche("pg_qualstats", 3);
 #else
 		RequestAddinLWLocks(2);
-#endif
+#endif		/* pg9.6+ */
+#endif		/* pg15- */
 	}
 	else
 		pgqs_backend_mode_startup();
 }
 
-void
-_PG_fini(void)
-{
-	/* Uninstall hooks. */
-	shmem_startup_hook = prev_shmem_startup_hook;
-	ExecutorStart_hook = prev_ExecutorStart;
-	ExecutorRun_hook = prev_ExecutorRun;
-	ExecutorFinish_hook = prev_ExecutorFinish;
-	ExecutorEnd_hook = prev_ExecutorEnd;
-}
-
 
 /*
  * Check that the sample ratio is in the correct interval
@@ -606,8 +615,13 @@ pgqs_ExecutorStart(QueryDesc *queryDesc, int eflags)
 #endif
 			)
 		{
+#if PG_VERSION_NUM >= 150000
+			query_is_sampled = (pg_prng_double(&pg_global_prng_state) <
+									pgqs_sample_rate);
+#else
 			query_is_sampled = (random() <= (MAX_RANDOM_VALUE *
 											 pgqs_sample_rate));
+#endif
 #if PG_VERSION_NUM >= 90600
 			pgqs_set_query_sampled(query_is_sampled);
 #endif
@@ -695,7 +709,6 @@ static void
 pgqs_ExecutorEnd(QueryDesc *queryDesc)
 {
 	pgqsQueryStringHashKey queryKey;
-	pgqsQueryStringEntry *queryEntry;
 	bool		found;
 
 	if ((pgqs || pgqs_backend) && pgqs_enabled && pgqs_is_query_sampled()
@@ -717,7 +730,6 @@ pgqs_ExecutorEnd(QueryDesc *queryDesc)
 	{
 		HASHCTL		info;
 		pgqsEntry  *localentry;
-		pgqsEntry  *newEntry;
 		HASH_SEQ_STATUS local_hash_seq;
 		pgqsWalkerContext *context = palloc(sizeof(pgqsWalkerContext));
 
@@ -738,13 +750,14 @@ pgqs_ExecutorEnd(QueryDesc *queryDesc)
 			/* Lookup the hash table entry with a shared lock. */
 			PGQS_LWL_ACQUIRE(pgqs->querylock, LW_SHARED);
 
-			queryEntry = (pgqsQueryStringEntry *) hash_search_with_hash_value(pgqs_query_examples_hash, &queryKey,
-																			  context->queryId,
-																			  HASH_FIND, &found);
+			hash_search_with_hash_value(pgqs_query_examples_hash, &queryKey,
+										context->queryId,
+										HASH_FIND, &found);
 
 			/* Create the new entry if not present */
 			if (!found)
 			{
+				pgqsQueryStringEntry *queryEntry;
 				bool		excl_found;
 
 				/* Need exclusive lock to add a new hashtable entry - promote */
@@ -813,7 +826,8 @@ pgqs_ExecutorEnd(QueryDesc *queryDesc)
 			hash_seq_init(&local_hash_seq, pgqs_localhash);
 			while ((localentry = hash_seq_search(&local_hash_seq)) != NULL)
 			{
-				newEntry = (pgqsEntry *) hash_search(pgqs_hash, &localentry->key,
+				pgqsEntry *newEntry = (pgqsEntry *) hash_search(pgqs_hash,
+													 &localentry->key,
 													 HASH_ENTER, &found);
 
 				if (!found)
@@ -931,7 +945,7 @@ pgqs_entry_copy_raw(pgqsEntry *dest, pgqsEntry *src)
  * Also maintain min and max values.
  */
 static void
-pgqs_entry_err_estim(pgqsEntry *e, double err_estim[2], int64 occurences)
+pgqs_entry_err_estim(pgqsEntry *e, double *err_estim, int64 occurences)
 {
 	int		i;
 
@@ -1359,7 +1373,8 @@ pgqs_process_booltest(BooleanTest *expr, pgqsWalkerContext *context)
 																	   GetDatabaseEncoding(),
 																	   PG_UTF8);
 
-			strncpy(entry->constvalue, utf8const, strlen(utf8const));
+			Assert(strlen(utf8const) < PGQS_CONSTANT_SIZE);
+			strcpy(entry->constvalue, utf8const);
 		}
 		else
 			memset(entry->constvalue, 0, sizeof(char) * PGQS_CONSTANT_SIZE);
@@ -1808,6 +1823,20 @@ pgqs_backend_mode_startup(void)
 #endif
 }
 
+#if PG_VERSION_NUM >= 150000
+static void
+pgqs_shmem_request(void)
+{
+	if (prev_shmem_request_hook)
+		prev_shmem_request_hook();
+
+	Assert(!pgqs_backend);
+
+	RequestAddinShmemSpace(pgqs_memsize());
+	RequestNamedLWLockTranche("pg_qualstats", 3);
+}
+#endif
+
 static void
 pgqs_shmem_startup(void)
 {
@@ -2031,8 +2060,7 @@ pg_qualstats_common(PG_FUNCTION_ARGS, pgqsVersion api_version,
 	nulls = palloc0(sizeof(bool) * nb_columns);
 	while ((entry = hash_seq_search(&hash_seq)) != NULL)
 	{
-		int			i = 0, j;
-		double		stddev_estim[2];
+		int			i = 0;
 
 		memset(values, 0, sizeof(Datum) * nb_columns);
 		memset(nulls, 0, sizeof(bool) * nb_columns);
@@ -2078,25 +2106,31 @@ pg_qualstats_common(PG_FUNCTION_ARGS, pgqsVersion api_version,
 
 		if (api_version >= PGQS_V2_0)
 		{
+			int		j;
+
 			for (j = 0; j < 2; j++)
+			{
+				double	stddev_estim;
+
+				if (j == PGQS_RATIO)	/* min/max ratio are double precision */
 				{
-					if (j == PGQS_RATIO)	/* min/max ratio are double precision */
-					{
-						values[i++] = Float8GetDatum(entry->min_err_estim[j]);
-						values[i++] = Float8GetDatum(entry->max_err_estim[j]);
-					}
-					else				/* min/max num are bigint */
-					{
-						values[i++] = Int64GetDatum(entry->min_err_estim[j]);
-						values[i++] = Int64GetDatum(entry->max_err_estim[j]);
-					}
-					values[i++] = Float8GetDatum(entry->mean_err_estim[j]);
-					if (entry->occurences > 1)
-						stddev_estim[j] = sqrt(entry->sum_err_estim[j] / entry->occurences);
-					else
-						stddev_estim[j] = 0.0;
-					values[i++] = Float8GetDatumFast(stddev_estim[j]);
+					values[i++] = Float8GetDatum(entry->min_err_estim[j]);
+					values[i++] = Float8GetDatum(entry->max_err_estim[j]);
 				}
+				else				/* min/max num are bigint */
+				{
+					values[i++] = Int64GetDatum(entry->min_err_estim[j]);
+					values[i++] = Int64GetDatum(entry->max_err_estim[j]);
+				}
+				values[i++] = Float8GetDatum(entry->mean_err_estim[j]);
+
+				if (entry->occurences > 1)
+					stddev_estim = sqrt(entry->sum_err_estim[j] / entry->occurences);
+				else
+					stddev_estim = 0.0;
+
+				values[i++] = Float8GetDatumFast(stddev_estim);
+			}
 		}
 
 		if (entry->position == -1)
@@ -2295,12 +2329,22 @@ pgqs_set_planstates(PlanState *planstate, pgqsWalkerContext *context)
 	context->inner_planstate = NULL;
 	context->planstate = planstate;
 	if (IsA(planstate, AppendState))
-		context->outer_planstate = ((AppendState *) planstate)->appendplans[0];
+	{
+		AppendState * appendstate = (AppendState *) planstate;
+		if (appendstate->as_nplans > 0)
+			context->outer_planstate = appendstate->appendplans[0];
+	}
 	else if (IsA(planstate, MergeAppendState))
-		context->outer_planstate = ((MergeAppendState *) planstate)->mergeplans[0];
+	{
+		MergeAppendState * mergeappendstate = (MergeAppendState *) planstate;
+		if (mergeappendstate->ms_nplans > 0)
+			context->outer_planstate = mergeappendstate->mergeplans[0];
+	}
 #if PG_VERSION_NUM < 140000
 	else if (IsA(planstate, ModifyTableState))
+	{
 		context->outer_planstate = ((ModifyTableState *) planstate)->mt_plans[0];
+	}
 #endif
 	else
 		context->outer_planstate = outerPlanState(planstate);
@@ -2422,15 +2466,53 @@ pgqs_memsize(void)
 static Size
 pgqs_sampled_array_size(void)
 {
+	int _maxbackends;
+
+#if PG_VERSION_NUM >= 150000
+	Assert(MaxBackends > 0);
+	_maxbackends = MaxBackends;
+#else
+	int32 _autovac_max_workers;
+	int32 _max_wal_senders;
+	const char *guc_string;
+
+	/*
+	 * autovacuum_max_workers and max_wal_senders aren't declared as
+	 * PGDLLIMPORT in pg15- versions, so retrieve them using GetConfigOption to
+	 * allow compilation on Windows.
+	 */
+	guc_string = GetConfigOption("autovacuum_max_workers",
+											 false, true);
+
+	_autovac_max_workers = pg_atoi(guc_string, 4, 0);
+	Assert(_autovac_max_workers >= 1 && _autovac_max_workers <= MAX_BACKENDS);
+
+	guc_string = GetConfigOption("max_wal_senders",
+											 false, true);
+
+	_max_wal_senders = pg_atoi(guc_string, 4, 0);
+	Assert(_max_wal_senders >= 0 && _max_wal_senders <= MAX_BACKENDS);
+
 	/*
 	 * Parallel workers need to be sampled if their original query is also
 	 * sampled.  We store in shared mem the sample state for each query,
 	 * identified by their BackendId.  If need room for all possible backends,
-	 * plus autovacuum launcher and workers, plus bg workers and an extra one
-	 * since BackendId numerotation starts at 1.
+	 * plus autovacuum launcher and workers, plus bg workers.
 	 */
-	return (sizeof(bool) * (MaxConnections + autovacuum_max_workers + 1
-							+ max_worker_processes + 1));
+	_maxbackends = MaxConnections + _autovac_max_workers + 1
+							+ max_worker_processes
+#if PG_VERSION_NUM >= 120000
+							/*
+							 * Starting with pg12, max_wal_senders isn't part
+							 * of max_connections anymore
+							 */
+							+ _max_wal_senders
+#endif		/* pg12+ */
+							+ 1;
+#endif		/* pg15- */
+
+	/* We need an extra value since BackendId numerotationn starts at 1. */
+	return (sizeof(bool) * (_maxbackends + 1));
 }
 #endif
 
